@@ -1,6 +1,5 @@
-/*
-   GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2009, 2010 Free Software Foundation, Inc.
+/* GNU Mailutils -- a suite of utilities for electronic mail
+   Copyright (C) 2009-2020 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -13,10 +12,8 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General
-   Public License along with this library; if not, write to the
-   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-   Boston, MA 02110-1301 USA
-*/
+   Public License along with this library.  If not, see
+   <http://www.gnu.org/licenses/>. */
 
 #include "libmu_py.h"
 
@@ -32,7 +29,7 @@ api_strerror (PyObject *self, PyObject *args)
     return NULL;
 
   str = (char *)mu_strerror (status);
-  return _ro (PyString_FromString (str));
+  return _ro (PyUnicode_FromString (str));
 }
 
 static PyMethodDef methods[] = {
@@ -43,8 +40,16 @@ static PyMethodDef methods[] = {
   { NULL, NULL, 0, NULL }
 };
 
+static struct PyModuleDef moduledef = {
+  PyModuleDef_HEAD_INIT,
+  PY_MODULE,
+  NULL,
+  -1,
+  methods
+};
+
 void
 _mu_py_attach_error (void)
 {
-  _mu_py_attach_module (PY_MODULE, methods);
+  _mu_py_attach_module (&moduledef);
 }
