@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999-2019 Free Software Foundation, Inc.
+   Copyright (C) 1999-2020 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -52,6 +52,135 @@ is_number (const char *s)
   return result;
 }
 
+static int
+match_hdr (mu_header_t hdr, void *pattern)
+{
+  int rc;
+  mu_iterator_t itr;
+  int result = 1;
+  
+  rc = mu_header_get_iterator (hdr, &itr);
+  if (rc)
+    {
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_header_get_iterator", NULL, rc);
+      return 1;
+    }
+  
+  for (mu_iterator_first (itr); result && !mu_iterator_is_done (itr);
+       mu_iterator_next (itr))
+    {
+      const char *value;
+      char *tmp, *s;
+      
+      mu_iterator_current (itr, (void**)&value);
+      tmp = strdup (value);
+      if (!tmp)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "strdup", NULL, errno);
+	  continue;
+	}
+      rc = mu_string_unfold (tmp, NULL);
+      if (rc)
+	{
+	  free (tmp);
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_string_unfold", NULL, rc);
+	  continue;
+	}
+      rc = mu_rfc2047_decode ("utf-8", tmp, &s);
+      free (tmp);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_rfc2047_decode", NULL, rc);
+	  continue;
+	}	  
+      tmp = s;
+      if (pattern_match (pattern, tmp))
+	result = 0;
+      free (tmp);
+    }
+  mu_iterator_destroy (&itr);
+  return result;
+}
+
+static int
+matchmsgpart (mu_message_t msg, void *pattern)
+{
+  mu_header_t hdr;
+  mu_stream_t str;
+  char *buf = NULL;
+  size_t bufsize = 0;
+  size_t n;
+  int result = 1;
+  int rc;
+  
+  /* Get the headers. */
+  rc = mu_message_get_header (msg, &hdr);
+  if (rc)
+    {
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_message_get_header", NULL, rc);
+      return 1;
+    }
+  if (match_hdr (hdr, pattern) == 0)
+    return 0;
+
+  if (message_body_stream (msg, 0, "utf-8", &str))
+    return 1;
+
+  /* Look for a matching line. */
+  while ((rc = mu_stream_getline (str, &buf, &bufsize, &n)) == 0 &&
+	 n > 0)
+    {
+      buf[n-1] = 0;
+      if (pattern_match (pattern, buf))
+	{
+	  result = 0;
+	  break;
+	}
+    }
+
+  free (buf);
+  mu_stream_destroy (&str);
+  return result;
+}
+
+static int
+matchmsg (mu_message_t msg, void *pattern)
+{
+  int rc;
+  mu_header_t hdr;
+  mu_iterator_t itr;
+  int result = 1;
+
+  mu_message_get_header (msg, &hdr);
+  if (match_hdr (hdr, pattern) == 0)
+    return 0;
+  rc = mu_message_get_iterator (msg, &itr);
+  if (rc)
+    {
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_message_get_iterator", NULL, rc);
+      return 0;
+    }
+  for (mu_iterator_first (itr); !mu_iterator_is_done (itr);
+       mu_iterator_next (itr))
+    {
+      mu_message_t partmsg;
+      
+      rc = mu_iterator_current (itr, (void**)&partmsg);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_iterator_current", NULL, rc);
+	  continue;
+	}
+      if (matchmsgpart (partmsg, pattern) == 0)
+	{
+	  result = 0;
+	  break;
+	}
+    }
+  mu_iterator_destroy (&itr);
+  return result;
+}
+
 /*
   According to ELM readmsg(1):
 
@@ -112,37 +241,22 @@ msglist (mu_mailbox_t mbox, int show_all, int argc, char **argv,
       else if (!is_number (argv[i]))
 	{
 	  size_t j;
-	  int found = 0;
+	  void *pat = pattern_init (argv[i]);
+	  if (!pat)
+	    continue;
 	  for (j = 1; j <= total; j++)
 	    {
-	      int status;
-	      char buf[128];
-	      size_t len = 0;
 	      mu_message_t msg = NULL;
-	      mu_stream_t stream = NULL;
 
 	      mu_mailbox_get_message (mbox, j, &msg);
-	      status = mu_message_get_streamref (msg, &stream);
-	      if (status)
-		mu_error (_("cannot read message: %s"),
-			  mu_strerror (status));
-	      else
+	      if (matchmsg (msg, pat) == 0)
 		{
-		  while (mu_stream_readline (stream, buf, sizeof buf, &len) == 0
-			 && len > 0)
-		    {
-		      if (strstr (buf, argv[i]) != NULL)
-			{
-			  addset (set, n, j);
-			  found = 1;
-			  break;
-			}
-		    }
-		  mu_stream_destroy (&stream);
+		  addset (set, n, j);
+		  if (!show_all)
+		    break;
 		}
-	      if (found && !show_all)
-		break;
 	    }
+	  pattern_free (pat);
 	}
       else if (mu_isdigit (argv[i][0]))
 	{
